# Generated by Django 2.2.16 on 2020-11-09 16:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('organizations', '0001_squashed_0007_historicalorganization'),
    ]

    operations = [
        migrations.AlterField(
            model_name='historicalorganization',
            name='short_name',
            field=models.CharField(db_index=True, help_text='Unique, short string identifier for organization. Please do not use spaces or special characters. Only allowed special characters are period (.), hyphen (-) and underscore (_).', max_length=255, verbose_name='Short Name'),
        ),
        migrations.AlterField(
            model_name='organization',
            name='short_name',
            field=models.CharField(help_text='Unique, short string identifier for organization. Please do not use spaces or special characters. Only allowed special characters are period (.), hyphen (-) and underscore (_).', max_length=255, unique=True, verbose_name='Short Name'),
        ),
    ]
